#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "schoenhage.h"
#include "hgcd.h"
#include "trace.h"

#define ASSERT(x) \
do { if (!(x)) abort(); } while (0)

#define trace_hgcd(tag, H)						\
  gmp_printf("%s: [ %Nd, %Nd\n"						\
	     "  %Nd, %Nd ]\n",						\
	     (tag),							\
	     (H)->uvp[0], (H)->uvsize[0], (H)->uvp[1], (H)->uvsize[1],	\
	     (H)->uvp[2], (H)->uvsize[2], (H)->uvp[3], (H)->uvsize[3])

static int
hgcd_schoenhage_equalp(struct hgcd *hgcd,
		       struct schoenhage_array *R)
{
  unsigned i, j;
  for (i = 0; i<2; i++)
    for (j = 0; j<2; j++)
      {
	unsigned s;
	if (R->m[i][j]->_mp_size != hgcd->uvsize[2*i+j])
	  {
	    trace("Element %d,%d differ in size\n", i, j);
	    return 0;
	  }
	s = hgcd->uvsize[2*i+j];
	if (s < 0) s = -s;
	
	while(s--)
	  if (R->m[i][j]->_mp_d[s] != hgcd->uvp[2*i+j][s])
	    {
	      trace("Element %d,%d differ in limb %d\n", i, j, s);
	      return 0;
	    }
      }
  return 1;
}

#if 0
static void
try_egcd(unsigned size, unsigned x, unsigned y)
{
  struct schoenhage_array R;
  struct schoenhage_array S;
  unsigned a, b;
  mpz_t s;
  mpz_t t;
  
  mpz_init(s);
  mpz_init(t);
  
  schoenhage_array_init(&R);
  schoenhage_array_identity(&R);

  a = x; b = y;

  for (;;)
    {
      unsigned q = a / b;
      unsigned r = a % b;

      trace("try_egcd: a = %d, b = %d, q = %d, r = %d\n",
	    a, b, q, r);
      
      schoenhage_array_step_ui(&R, q);
      
      trace(" R = [ %Zd, %Zd ; %Zd, %Zd ]\n",
	    R.m[0][0], R.m[0][1], R.m[1][0], R.m[1][1]);

      /* Stop here, to match what egcd does */
      if (r < 2)
	break;
      
      mpz_mul_si(s, R.m[0][0], x);
      mpz_mul_si(t, R.m[0][1], y);
      mpz_add(s, s, t);

      ASSERT(!mpz_cmp_si(s, b));

      a = b; b = r;
    }

  schoenhage_array_init(&S);
  egcd(&S, size, x, y);

  trace_array(" S", &S);
  
  ASSERT(size || schoenhage_array_equal(&R, &S));
  
  schoenhage_array_clear(&R);
  schoenhage_array_clear(&S);
  mpz_clear(s);
  mpz_clear(t);
}
#endif

static void
try_hgcd(mpz_t a, mpz_t b)
{
  struct schoenhage_array R;
  struct schoenhage_array S;
  struct hgcd H;

  schoenhage_array_init(&R);
  schoenhage_array_init(&S);
  
  trace("try_hgcd: a = %Zd, b = %Zd\n", a, b);
  hgcd_slow(&R, a, b);
  check_hgcd(&R, a, b);
  hgcd(&S, a, b);
  check_hgcd(&S, a, b);

  trace_array(" R", &R);
  trace_array(" S", &S);

  if (!schoenhage_array_equal(&R, &S))
    trace("try_hgcd: Mismatch\n");
  /* ASSERT(schoenhage_array_equal(&R, &S)); */

  if (mpz_size(a) >= 2)
    {
      mp_size_t scratch = mpn_hgcd_itch(mpz_size(a));
      mpn_hgcd_init(&H, mpz_size(a),
		    alloca( sizeof(mp_limb_t) * scratch), scratch);
      
      /* Ordinary gcd needs an extra limb. */
      mpz_realloc(a, mpz_size(a) + 1);
      mpz_realloc(b, mpz_size(b) + 1);
      
      H.ap = a->_mp_d;
      H.asize = a->_mp_size;
      H.bp = b->_mp_d;
      H.bsize = b->_mp_size;
      
      mpn_lehmer_hgcd(&H); /* Destroys contents of a and b */
      /* Should give a precise result, matching hgcd_slow: R. No, not
       * really, as they use different bases. */
      
      trace_hgcd("H", &H);
      mpn_check_hgcd(&H, a->_mp_size);
      
      /* ASSERT(hgcd_schoenhage_equalp(&H, &R)); */
    }
  schoenhage_array_clear(&R);
  schoenhage_array_clear(&S);
}

static void
try_hgcd_str(const char *sa, const char *sb)
{  
  mpz_t a;
  mpz_t b;

  mpz_init(a);
  mpz_init(b);

  mpz_set_str(a, sa, 10);
  mpz_set_str(b, sb, 10);

  try_hgcd(a, b);
  
  mpz_clear(a);
  mpz_clear(b);
}

static void
try_hgcd_fib(unsigned n)
{
  mpz_t a;
  mpz_t b;

  mpz_init(a);
  mpz_init(b);

  mpz_fib2_ui(a, b, n);

  try_hgcd(a,b);

  mpz_clear(a);
  mpz_clear(b);
}

static void
try_fgcd(unsigned n)
{
  mpz_t a;
  mpz_t b;
  mpz_t r;
  
  mpz_init(a);
  mpz_init(b);
  mpz_init(r);

  trace("\ntry_fgcd: n = %d\n", n);
  
  mpz_fib2_ui(a, b, n);
  schoenhage(r, a, b);
  trace("schoenhage(%Zd, %Zd) = %Zd\n",
	a, b, r);
  ASSERT(mpz_cmp_ui(r, 1) == 0);
  
  /* Gcd needs an extra limb. */
  mpz_realloc(a, mpz_size(a) + 1);
  mpz_realloc(b, mpz_size(b) + 1);
  mpz_realloc(r, mpz_size(a));
  
  r->_mp_size = mpn_fgcd(r->_mp_d,
			 a->_mp_d, a->_mp_size,
			 b->_mp_d, b->_mp_size);

  ASSERT(mpz_cmp_ui(r, 1) == 0);
  
  mpz_clear(a);
  mpz_clear(b);
  mpz_clear(r);  
}

#define RANDOM(r, state, size) do {		\
  mpz_urandomb((r), (state), (size));		\
  mpz_setbit(r, (size)-1);			\
} while (0)

static void
try_random(void)
{
  unsigned size;
  
  gmp_randstate_t state;
  time_t now;

  /* now = time(NULL); */
  now = 1055513922;

  trace("try_random: Seeding with %ld\n", (long) now);
  
  gmp_randinit_default(state);
  gmp_randseed_ui(state, now);

  for (size = 10; size < 1000; size += 3)
    {
      mpz_t a;
      mpz_t b;
      mpz_t c;
      mpz_t g;
      mpz_t t;

      mpz_init(a);
      mpz_init(b);
      mpz_init(c);
      mpz_init(g);
      mpz_init(t);

      RANDOM(a, state, size + 17);
      RANDOM(b, state, size + 47);
      RANDOM(t, state, size);

      mpz_mul(a, a, t);
      mpz_mul(b, b, t);
      mpz_gcd(g, a, b);

      trace("try_random: size = %d, gcd(%Zd, %Zd) = %Zd\n",
	    size, a, b, g);

      schoenhage(t, a, b);
      if (mpz_cmp(t, g))
	{
	  trace("try_random: Incorrect result: %Zd\n");
	  abort();
	}

      /* Gcd needs an extra limb. */
      mpz_realloc(a, mpz_size(a) + 1);
      mpz_realloc(b, mpz_size(b) + 1);
      mpz_realloc(t, mpz_size(a));

      t->_mp_size = mpn_fgcd(t->_mp_d,
			     a->_mp_d, a->_mp_size,
			     b->_mp_d, b->_mp_size);
      
      mpz_clear(a);
      mpz_clear(b);
      mpz_clear(c);
      mpz_clear(g);
      mpz_clear(t);
    }
  
      
  gmp_randclear(state);
}

int
main(int argc, char **argv)
{
  unsigned n;
  unsigned i;

  mpz_t a;
  mpz_t b;
  mpz_t r;
  
  mpz_init(a);
  mpz_init(b);
  mpz_init(r);

  n = 0;
  if (argc > 1)
    n = atoi(argv[1]);
  
  /* normalize */
  mpz_set_si(a, -7);
  mpz_set_si(b, 3);
  
  ASSERT(normalize(r, a, b));
  ASSERT(!mpz_cmp_si(a, 7));
  ASSERT(!mpz_cmp_si(b, 3));

  mpz_set_si(a, 0);
  ASSERT(!normalize(r, a, b));
  ASSERT(!mpz_cmp_si(r, 3));

#if 0
  /* schoenhage_array_step_ui */
  try_egcd(0, 268, 87);
  try_egcd(0, 270, 87);
  /* Problematic case for gcd(fib(100), fib(99)) */
  try_egcd(8, 87833, 6075);
#endif
  
  for (i = 6; i<17; i++)
    try_hgcd_fib(i);

  /* FIXME: There seems to be a problem with fib 109, which uses
     3-limb values. */
  try_hgcd_str("133304756", "12777598");
  try_hgcd_str("4368130264137", "418696360162");

  /* Problematic case, because one lehmer step reduces the numbers too much,
     to 17019354528843361985, 10518539562755788252 */
  
  try_hgcd_str("69883482110095079835122363461",
	       "43190367196233980458636766295");

  try_hgcd_str("30019063369843142", "9258668778823265"); /* !!! */
  
  try_hgcd_str("133304756", "12777598");
  try_hgcd_str("223659451", "68982457");
  try_hgcd_str("8259214276", "5104443020");
  try_hgcd_str("70946110418820658019", "43846831676944464992");
  try_hgcd_str("14901987450837196642636732651392",
	       "11108588010711814328108699732800");
  try_hgcd_str("1225680876558579559649902923502714694035764",
	       "812484106832574967003225854678869801985244");
  
  if (n)
    try_fgcd(n);
  else
    {
      for (i = 1; i<10; i++)
	try_fgcd(i);
  
      for( ; i < 100000; i *= 10)
	try_fgcd(i);
    }

  try_random();
  
  mpz_clear(a);
  mpz_clear(b);
  mpz_clear(r);

  return EXIT_SUCCESS;
}
